# Libraries that are needed in this script
import numpy as np

################################################################################
#
# Definition of the LeastSquares class. This class represents a least squares
# problem in which a matrix A and a vector y are given. When we instantiate an
# object of this class we pass it the matrix A and the vector y for storage in
# respective attributes of the class.
#
# The class has two attributes
#
#     A = The least squares matrix
#     y = The least squares vector 
#
# The class has two methods:
#
#     __init__ : Called at initialization. Receives A as a parameter.
#     solve    : Called to return the solution ot the least squares problem.
#
class LeastSquares():

    def __init__(self, A, y):
        self.A = A
        self.y = y        

    def solve(self):
        A = self.A
        y = self.y
        AT  = np.transpose(A)       # A transpose
        ATA = np.matmul(A,AT)       # A transpose * A
        iATA = np.linalg.inv(ATA)   # inverse(A transpose * A)
        iATAAT = np.matmul(AT,iATA) # inv(A transpose * A) * (A transpose)
        x = np.matmul(y,iATAAT)     # inv(A transpose * A) * (A transpose) * y
        return x

################################################################################
#
# Definition of the GaussianLeastSquares class. This class represents a least
# squares problem in which the entries of the matrix A and the vector y are chosen
# from a Gaussian distribution with mean 0 and variance 1. This could be useful if,
# say, you decide to study the properties of least squares problem for
# matrices and vectors of this type.
#
# The class has two attributes
#
#     A = The least squares matrix. Randomly generated to have m rows and
#         n columns. Entries are drawn from Gaussian distribution with
#         mean 0 and variance 1.
#     y = The least squares matrix. Randomly generated to have m rows and
#         n columns. Entries are drawn from Gaussian distribution with
#         mean 0 and variance 1.
#
# The class has two methods:
#
#     __init__ : Called at initialization. Receives dimensions m and n
#                as parameters.
#     solve    : Called to return the solution ot the least squares problem.
# 
# The method solve is not rewritten. It is is inheirted from the parent class.
#
class GaussianLeastSquares(LeastSquares):

    def __init__(self, m, n):
        self.A = np.random.normal(0, 1, size=(n, m))
        self.y = np.random.normal(0, 1, size=(n, m))        

################################################################################
#    
# A function to print a matrix (or vector). It formats the numbers and prints
# the matrix in a way that makes it easy to read. It is not important. 
#
# The arguments we pass to the function are:
#
#     A = The matrix to be printed
#     nr_decimals = the number of decimals to be printed. It is an optional
#                   argument. Set to 2 by default

def print_matrix(A, nr_decimals = 2):

    # Determine the number of digits in the largest number in the matrix and use
    # it to specify the number format

    nr_digits = np.maximum(np.floor(np.log10(np.amax(np.abs(A)))),0) + 1
    nr_digits = nr_digits + nr_decimals + 3
    nr_digits = "{0:1.0f}".format(nr_digits)
    number_format = "{0: " + nr_digits + "." + str(nr_decimals) + "f}"
    
    # Determine matrix size
    n = len(A)
    m = len(A[0])

    # Sweep through rows
    for l in range(m):
        value = " "

        # Sweep through columns
        for k in range(n):

            # ccncatenate entries to create row printout
            value = value + " " + number_format.format(A[k,l])

        # Print row
        print( value )
